/*
    Copyright 2005-2012 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>
#include <time.h>

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef int32_t     i32;
typedef uint32_t    u32;
typedef uint64_t    u64;



// #define BITS32   // uncomment to enable 32 bits
// #define DEBUG    // uncomment to enable the step-by-step visualization



#define BITS64
#ifdef BITS32
    #undef BITS64
#endif

#ifdef BITS64
    #define NTI     u64
    #define NTF     double
    #define NTSZ    64
#else
    #define NTI     u32
    #define NTF     float
    #define NTSZ    32
#endif


#define VER         "0.1.5"
#define READSZ      512
#define MOV(X,Y)    memmove(X, Y, slen - (Y - buff));
#define NTCHRSZ     (NTSZ >> 3)
#define SHOWBASETMP 20



void help(void);
int findop(int c);
NTI rol(NTI n1, NTI n2);
NTI ror(NTI n1, NTI n2);
NTI ip2num(char *ip);
char *num2ip(NTI num);
NTI bitswap(NTI n1, NTI n2);
NTI byteswap(NTI n1, NTI n2);
NTI mysqrt(NTI num);
NTI radix(NTI n1, NTI n2);
NTI power(NTI n1, NTI n2);
NTI get_num(char *p, NTI left);
int mytolower(char *data);
NTI readdouble(char *data);
NTI readbase(char *data, int size);
NTI readtime(char *data);
u8 *time64_to_strtime(NTI num);
char *showtime(NTI num);
char *showchar(NTI num);
char *showbase(NTI num, int size, int sign);



static const u8 b64_table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const u8 b32_table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
static const u8 hex_table[] = "0123456789abcdef";



int main(int argc, char *argv[]) {
    NTI     left,
            right,
            tot;
    NTF     ntftmp;
    double  totdb;
    float   totfl;
    int     op,
            len,
            slen;
    char    buff[READSZ * 9],   // should be enough
            tmp[NTSZ + 1],
            *data,
            *pl,
            *pr,
            *p,
            *l,
            *special;

    setbuf(stdout, NULL);

    printf("\n"
        "Calcc "VER" (%d bits version)\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", NTSZ);

    if(argc > 1) {
        p = buff;
        l = buff + READSZ - 1;
        data = argv[1];
        while(*data && (p < l)) {
            *p++ = *data++;
        }
        *p = 0;
        goto do_it;
    }

    while(fgets(buff, READSZ, stdin)) {
        for(p = buff; *p && (*p != '\r') && (*p != '\n'); p++);
        *p = 0;

do_it:
        //mytolower(buff);  don't use it or "" will not work

        for(p = data = buff; *p; p++) {
            if(*p == '\"') *p = '\'';
            if(*p == '\'') {
                for(l = p; *l; l++) {
                    if(*l == '\"') *l = '\'';
                    *data++ = *l;
                    if((l != p) && (*l == '\'')) break;
                }
                p = l;
                continue;
            }

            *p = tolower(*p);
            if((*p == '\r') || (*p == '\n')) {
                break;
            } else if(*p <= ' ') {
                continue;
            } else if((*p == '[') || (*p == '{')) {
                *data++ = '(';
            } else if((*p == ']') || (*p == '}')) {
                *data++ = ')';
            } else if((*p == '<') && (p[1] == '<')) {
                continue;
            } else if((*p == '>') && (p[1] == '>')) {
                continue;
            } else if((*p == '&') && (p[1] == '&')) {
                continue;
            } else if((*p == '|') && (p[1] == '|')) {
                continue;
            } else {
                *data++ = *p;
            }
        }
        *data = 0;
        slen = (data - buff) + 1;   // NULL included!!!
        if(!strcmp(buff, "?") || !strcmp(buff, "help")) {   // they are already lower
            help();
            *buff = 0;
        } else if(!strcmp(buff, "q") || !strcmp(buff, "quit")) {   // they are already lower
            exit(0);
            *buff = 0;
        }

        for(;;) {
            data = buff;

            #ifdef DEBUG
            printf("DEBUG: %s\n", data);
            #endif

            special = NULL;
            if(*data == '\'') {
                special = data;
                for(++data; *data && (*data != '\''); data++);
                if(*data) data++;
            }

            pr = strchr(data, ')');
            if(pr) {
                for(pl = pr; (pl >= data) && (*pl != '('); pl--);
                data = pl + 1;
                if(data == buff) {
                    printf("- no opened parenthesis found\n");
                    *data = 0;
                    break;
                }

                for(p = data; !findop(*p); p++);
                if(p == pr) {   // find (num)
                    memmove(pl, pl + 1, pr - (pl + 1));
                    MOV(pr - 1, pr + 1);
                    slen -= 2;
                    continue;
                }
                *pr = 0;
            } else if((pl = strchr(data, '('))) {
                printf("- no closed parenthesis found\n");
                *data = 0;
                break;
            }

            #define BLA(x)  if((p = strchr(data, x))) { op = x;
            #define BLAH(x) BLA(x) }

            // priority based only on the operator
            // I don't consider the left/right position at the moment
            // the user MUST ever use the parenthesis
            BLA('~')
                MOV(p + 1, p);
                slen++;
                if(!pr) pr = p;
                pr++;
                *p++ = '0';
            } else BLA('!')
                MOV(p + 1, p);
                slen++;
                if(!pr) pr = p;
                pr++;
                *p++ = '0';
            }
            else BLAH('<')
            else BLAH('>')
            else BLAH('l')
            else BLAH('r')
            else BLAH('&')
            else BLAH('^')
            else BLAH('|')
            else BLAH('p')
            else BLAH('v')
            else BLAH('s')
            else BLAH('w')
            else BLAH('*')
            else BLAH('/')
            else BLAH('%')
            else BLA('-')
                if(p[1] == '-') {
                    l = p + 2;
                    findop(*l) ? (p[1] = '1') : (*p = '1');
                }
            }
            else BLA('+')
                if(p[1] == '+') {
                    l = p + 2;
                    findop(*l) ? (p[1] = '1') : (*p = '1');
                }
            }
            else BLAH('=')
            else break;

            #undef BLA
            #undef BLAH

            for(l = p - 1; (l >= data) && !findop(*l); l--);

            if(special) l = special - 1;
            left  = get_num(++l, 1);
            right = get_num(++p, left);

            switch(op) {
                case '~': tot = right ^ ((NTI)-1);      break;
                case '!': tot = right ? 0 : 1;          break;
                case '<': { // x86 fix!
                    tot = ((right >= NTSZ) ? 0 : (left << right));
                    break;
                }
                case '>': { // x86 fix!
                    tot = ((right >= NTSZ) ? 0 : (left >> right));
                    break;
                }
                case 'l': tot = rol(left, right);       break;
                case 'r': tot = ror(left, right);       break;
                case '^': tot = left ^ right;           break;
                case '&': tot = left & right;           break;
                case '|': tot = left | right;           break;
                case 'p': tot = power(left, right);     break;
                case 'v': tot = radix(left, right);     break;
                case 's': tot = byteswap(left, right);  break;
                case 'w': tot = bitswap(left, right);   break;
                case '*': tot = left * right;           break;
                case '/': {
                    if(!right) {
                        printf("division by zero\n");
                        tot = 0;
                    } else {
                        tot = left / right;
                    }
                    break;
                }
                case '%': {
                    if(!right) {
                        printf("division by zero\n");
                        tot = 0;
                    } else {
                        tot = left % right;
                    }
                    break;
                }
                case '+': tot = left + right;           break;
                case '-': tot = left - right;           break;
                case '=': tot = right;                  break;
                default:  tot = 0;                      break;
            }

            if(pr) *pr = ')';

            while(!findop(*p)) p++;

            len = sprintf(tmp, "h%s", showbase(tot, 16, 0));
            if(len != (p - l)) {
                MOV(l + len, p);
                slen += (l + len) - p;
            }
            memcpy(l, tmp, len);
        }

        if(!*buff) continue;

        tot = get_num(buff, 1);

        ntftmp = 0;
        memcpy(&ntftmp, &tot, sizeof(ntftmp));
        totdb = (double)ntftmp;
        totfl = (float)ntftmp;

        printf(
//#ifdef BITS64
            " %-20s   %-16s   %s\n"
            " %-64s\n"
            " %-21s  %-16s   %-.19e\n"
            " %-32s          %s\n"
            " %-14s         %-12s       %s\n"
            " %-21.8f  %-21.16g\n"
            " %s\n"
/*
#else
            " %-10s   %-8s   %-11s        %-32s\n"
            " %-11s  %-8s   %-.11e %-16s   %s\n"
            " %-14s          %-12s       %s\n"
            " %-21.8f   %-11.16g\n"
            " %s\n"
#endif
*/
            "\n",
            showbase(tot, 10, 0), showbase(tot, 16, 0), showbase(tot, 8, 0),
            showbase(tot, 2,  0),
            showbase(tot, 10, 1), showchar(tot), (NTF)tot,
            showbase(tot, 4,  0), num2ip(tot),
            showbase(tot, 32, 0), showbase(tot, 64, 0), showtime(tot),
            totfl, totdb, time64_to_strtime(tot));
    }

    return(0);
}



void help(void) {
    fputs("\n"
        "Input types:\n"
        "  0x = hexadecimal                     h  = hexadecimal\n"
        "  $  = hexadecimal                     o  = octal\n"
        "  b  = binary                          i  = IP address (v4 or v6)\n"
        "  q  = base_four                       c  = percentage\n"
        "  d  = day                             m  = month\n"
        "  y  = year                            t  = time hh:mm:ss (0 for current date)\n"
        "  ,. = float/double numbers (it's not possible to make operations on them)\n"
        "  \"' = char or string                     = decimal (default)\n"
        "\n"
        "Operators and priority order:\n"
        "  ~  = complement                      !  = not\n"
        "  << = shift left                      >> = shift right\n"
        "  l  = rotate left                     r  = rotate right\n"
        "  ^  = xor              &  = and       |  = or\n"
        "  p  = power                           v  = root\n"
        "  s  = byte swapping                   w  = bit swapping\n"
        "  *  = multiplication   /  = division  %  = modulus\n"
        "  +  = addition                        -  = substraction\n"
        "\n"
        "Outputs:\n"
#ifdef BITS64
        "  decimal          hexadecimal      octal\n"
        "  binary\n"
        "  signed_decimal   ascii            exponential\n"
        "  base_four                         IPv4\n"
        "  base32_rfc4648   base64_rfc4648   date/time\n"
        "  float(test)      double(test)\n"
        "  64bit_date/time\n"
#else
        "  decimal          hexadecimal      octal            binary\n"
        "  signed_decimal   ascii            exponential      base_four   IP_address\n"
        "  base32_rfc4648                    base64_rfc4648   date/time\n"
        "  float(test)      double(test)\n"
        "  64bit_date/time\n"
#endif
        "\n", stdout);
}



int findop(int c) {
    return(
        (c == '~') || (c == '!') ||
        (c == '<') || (c == '>') ||
        (c == 'l') || (c == 'r') ||
        (c == '^') || (c == '&') || (c == '|') ||
        (c == 'p') || (c == 'v') ||
        (c == 's') || (c == 'w') ||
        (c == '*') || (c == '/') || (c == '%') ||
        (c == '+') || (c == '-') ||
        (c == '=') ||
        (c == '(') || (c == ')') ||
        !c);
}



typedef struct {
    u16     Year;
    u16     Month;
    u16     Day;
    u16     Hour;
    u16     Minute;
    u16     Second;
    u16     Milliseconds;
    u16     Weekday;
} PTIME_FIELDS;



// from Wine source code
// http://source.winehq.org/source/dlls/ntdll/time.c#L115
void myRtlTimeToTimeFields(u64 *liTime, PTIME_FIELDS *TimeFields) {
    #define TICKSPERSEC        10000000
    #define TICKSPERMSEC       10000
    #define SECSPERDAY         86400
    #define SECSPERHOUR        3600
    #define SECSPERMIN         60
    //#define MINSPERHOUR        60
    //#define HOURSPERDAY        24
    #define EPOCHWEEKDAY       1  /* Jan 1, 1601 was Monday */
    #define DAYSPERWEEK        7
    //#define EPOCHYEAR          1601
    //#define DAYSPERNORMALYEAR  365
    //#define DAYSPERLEAPYEAR    366
    //#define MONSPERYEAR        12
    #define DAYSPERQUADRICENTENNIUM (365 * 400 + 97)
    //#define DAYSPERNORMALCENTURY (365 * 100 + 24)
    #define DAYSPERNORMALQUADRENNIUM (365 * 4 + 1)

    i32 SecondsInDay;
    u32 cleaps, years, yearday, months;
    u32 Days;
    u64 Time;

    /* Extract millisecond from time and convert time into seconds */
    TimeFields->Milliseconds =
        (u16) (( *liTime % TICKSPERSEC) / TICKSPERMSEC);
    Time = *liTime / TICKSPERSEC;

    /* The native version of RtlTimeToTimeFields does not take leap seconds
     * into account */

    /* Split the time into days and seconds within the day */
    Days = Time / SECSPERDAY;
    SecondsInDay = Time % SECSPERDAY;

    /* compute time of day */
    TimeFields->Hour = (u16) (SecondsInDay / SECSPERHOUR);
    SecondsInDay = SecondsInDay % SECSPERHOUR;
    TimeFields->Minute = (u16) (SecondsInDay / SECSPERMIN);
    TimeFields->Second = (u16) (SecondsInDay % SECSPERMIN);

    /* compute day of week */
    TimeFields->Weekday = (u16) ((EPOCHWEEKDAY + Days) % DAYSPERWEEK);

    /* compute year, month and day of month. */
    cleaps=( 3 * ((4 * Days + 1227) / DAYSPERQUADRICENTENNIUM) + 3 ) / 4;
    Days += 28188 + cleaps;
    years = (20 * Days - 2442) / (5 * DAYSPERNORMALQUADRENNIUM);
    yearday = Days - (years * DAYSPERNORMALQUADRENNIUM)/4;
    months = (64 * yearday) / 1959;
    /* the result is based on a year starting on March.
     * To convert take 12 from Januari and Februari and
     * increase the year by one. */
    if( months < 14 ) {
        TimeFields->Month = months - 1;
        TimeFields->Year = years + 1524;
    } else {
        TimeFields->Month = months - 13;
        TimeFields->Year = years + 1525;
    }
    /* calculation of day of month is based on the wonderful
     * sequence of INT( n * 30.6): it reproduces the 
     * 31-30-31-30-31-31 month lengths exactly for small n's */
    TimeFields->Day = yearday - (1959 * months) / 64 ;
}



u8 *time64_to_strtime(NTI num) {
    static u8   buff[32];
    static const char   *months[] = { "Jan","Feb","Mar","Apr","May","Jun", "Jul","Aug","Sep","Oct","Nov","Dec" };
    PTIME_FIELDS    tm;
    u64     tmp64;

    tmp64 = num;
    myRtlTimeToTimeFields(&tmp64, &tm);

    sprintf(
        buff,
        "%02d %s %d %02d:%02d:%02d",
        tm.Day,
        months[(tm.Month - 1) % 12],
        tm.Year,
        tm.Hour, tm.Minute, tm.Second);
    return(buff);
}



NTI rol(NTI n1, NTI n2) {
    return((n1 << n2) | (n1 >> (NTSZ - n2)));
}



NTI ror(NTI n1, NTI n2) {
    return((n1 >> n2) | (n1 << (NTSZ - n2)));
}



NTI ip2num(char *ip) {
    NTI     out = 0;
    int     i,
            len;
    unsigned oct[6];

    for(i = 0; i < 6; i++) {
        oct[i] = 0;
    }

    len = sscanf(ip,
        "%u.%u.%u.%u.%u.%u",
        &oct[0], &oct[1], &oct[2], &oct[3], &oct[4], &oct[5]);

    for(i = 0; i < 6; i++) {
        oct[i] &= 0xff;  // needed!
    }

    for(i = 0; i < len; i++) {
        out = (out << 8) | oct[i];
    }
    return(out);
}



char *num2ip(NTI num) {
    static char out[24];

#ifdef BITS64
    if(num > (NTI)0xffffffff)    // IPv6
        sprintf(out, "IPv6 unsupported");   // unsupported because it can't handle that size
    else                    // IPv4
#endif
        sprintf(out,
            "%u.%u.%u.%u",
            (u8)((num >> 24) & 0xff),
            (u8)((num >> 16) & 0xff),
            (u8)((num >> 8)  & 0xff),
            (u8)(num         & 0xff));

    return(out);
}



NTI bitswap(NTI n1, NTI n2) {
    NTI     out,
            rem = 0;

    if(n2 < NTSZ) {
        rem = n1 & (((NTI)-1) ^ (((NTI)1 << n2) - 1));
    }

    for(out = 0; n2; n2--) {
        out = (out << 1) | (n1 & 1);
        n1 >>= 1;
    }
    return(out | rem);
}



NTI byteswap(NTI n1, NTI n2) {
    NTI     out,
            rem = 0;

    if(n2 < (NTSZ >> 3)) {
        rem = n1 & (((NTI)-1) ^ (((NTI)1 << (n2 << 3)) - 1));
    }

    for(out = 0; n2; n2--) {
        out = (out << 8) | (n1 & 0xff);
        n1 >>= 8;
    }
    return(out | rem);
}



    // the following function should be part of the
    // ffmpeg project but I have not verified it
NTI mysqrt(NTI num) {
     NTI    ret    = 0,
            ret_sq = 0,
            b;
     int    s;

     for(s = (NTSZ >> 1) - 1; s >= 0; s--) {
         b = ret_sq + ((NTI)1 << (s << 1)) + ((ret << s) << 1);
         if(b <= num) {
             ret_sq = b;
             ret += (NTI)1 << s;
         }
     }
     return(ret);
}



NTI radix(NTI n1, NTI n2) {
    NTI     i,
            old,    // due to the
            new;    // lack of bits

    if(!n1 || !n2) return(0);

    if(n2 == 2) return(mysqrt(n1)); // fast way

    for(i = old = 1; ; i <<= 1) {   // faster???
        new = power(i, n2);
        if((new > n1) || (new < old)) break;
        old = new;
    }

    for(i >>= 1; ; i++) {
        new = power(i, n2);
        if((new > n1) || (new < old)) break;
        old = new;
    }
    return(i - 1);
}



    // optimized from http://www.hackersdelight.org/HDcode/iexp.cc
NTI power(NTI n1, NTI n2) {
    NTI     out = 1;

    for(;;) {
        if(n2 & 1) out *= n1;
        n2 >>= 1;
        if(!n2) break;
        n1 *= n1;
    }
    return(out);
}



NTI get_num(char *p, NTI left) {
    NTI     num;
    int     dot;
    char    *l;

    if(findop(*p)) return(0);

    dot = 0;        // auto IP
    for(l = p; !findop(*l); l++) {
        if(*l == '.') {
            dot++;
            if(dot == 3) return(ip2num(p));
        }
    }
    if((dot == 1) || strchr(p, ',')) return(readdouble(p));

    if((*p == '0') && (p[1] == 'x')) p++;  // hex

    num = 0;    // compatibility
    switch(*p) {
        case 'x':
        case '$':
        case 'h': num = readbase(p + 1, 16);            break;
        case 'o': num = readbase(p + 1, 8);             break;
        case 'b': num = readbase(p + 1, 2);             break;
        case 'i': num = ip2num(p + 1);                  break;
        case 'q': num = readbase(p + 1, 4);             break;
        case 't': num = readtime(p + 1);                break;
        case 'd': num = readbase(p + 1, 10) * 86400;    break;
        case 'm': num = readbase(p + 1, 10) * 2629744;  break;  // 28,29,30 and 31
        case 'y': num = readbase(p + 1, 10) * 31556926; break;
        case 'c': {
            num = readbase(p + 1, 10) * left;
            num /= 100;
            break;
        }
        //case '\"':  // impossible because " gets replaced by '
        case '\'': {
            for(++p; *p && (*p != '\''); p++) {
                num <<= 8;
                num |= *p;
            }
            break;
        }
        default:  num = readbase(p, 10);                break;
    }
    return(num);
}



int mytolower(char *data) {
    char    *p;

    for(p = data; *p; p++) {
        *p = tolower(*p);
    }
    return(p - data);
}



NTI readdouble(char *data) {
    NTI     out;
    double  db;
    float   fl;

    fl = atof(data);
    db = atof(data);
    out = 0;
    if(sizeof(out) == sizeof(db)) {
        memcpy(&out, &db, sizeof(out));
    } else {
        memcpy(&out, &fl, sizeof(out));
    }
    return(out);
}



NTI readbase(char *data, int size) {
    NTI     num;
    int     sign = 0;
    u8      c,
            *p,
            *table;

    if((size > 0) && (size <= 16)) {
        table = (u8 *)hex_table;
    } else if(size == 32) {
        table = (u8 *)b32_table;
    } else if(size == 64) {
        table = (u8 *)b64_table;
    } else {
        return(0);
    }
    if(*data == '-') {  // useless in calcc but can useful in other programs
        sign = 1;
        data++;
    }
    for(num = 0; *data; data++) {
        c = *data;      // already tolower()
        if((c == 'x') && (size == 16)) continue;
        p = memchr(table, c, size);
        if(!p) break;
        num = (num * size) + (p - table);
    }
    if(sign) num = -num;
    return(num);
}



NTI readtime(char *data) {
    NTI     out = 0;
    int     ss = 0,
            mm = 0,
            hh = 0;

    sscanf(
        data,
        "%02d:%02d:%02d",
        &hh,
        &mm,
        &ss);

    // hh %= 24;    // maximum freedom of using bigger values
    // mm %= 60;
    // ss %= 60;
    out = (hh * 60 * 60) + (mm * 60) + ss;
    if(!out) out = time(NULL);
    return(out);
}



char *showtime(NTI num) {
    struct  tm  *tmx;
    time_t  datex;
    static  char    buff[32];
    static const char   *day[] =
            { "Sun","Mon","Tue","Wed","Thu","Fri","Sat" },
            *months[] =
            { "Jan","Feb","Mar","Apr","May","Jun", "Jul","Aug","Sep","Oct","Nov","Dec" };

    datex = num;
    tmx = gmtime(&datex);
    if(!tmx) return("none");
    sprintf(
        buff,
        "%s %02d %s %d %02d:%02d:%02d",
        day[tmx->tm_wday % 7],
        tmx->tm_mday,
        months[tmx->tm_mon % 12],
        1900 + tmx->tm_year,
        tmx->tm_hour, tmx->tm_min, tmx->tm_sec);
    return(buff);
}



char *showchar(NTI num) {
    int     i;
    static char out[NTCHRSZ + 1];
    char    *p;

    p = out;
    for(i = NTSZ - 8; i >= 0; i -= 8) {
        *p = (num >> i) & 0xff;
        if((p == out) && !*p) continue;
        if(*p < ' ') *p = '.';
        p++;
    }
    *p = 0;
    return(out);
}



char *showbase(NTI num, int size, int sign) {
    static char out[SHOWBASETMP][NTSZ + 1];
    static int  outp    = 0;
    int     i;
    u8      *o,
            *table;

    if((size > 0) && (size <= 16)) {
        table = (u8 *)hex_table;
    } else if(size == 32) {
        table = (u8 *)b32_table;
    } else if(size == 64) {
        table = (u8 *)b64_table;
    } else {
        return("");
    }
    o = out[outp % SHOWBASETMP];
    outp++;

    if(sign && (num >= ((NTI)1 << (NTSZ - 1)))) {
        num = -num;
    } else {
        sign = 0;
    }
    memset(o, '0', NTSZ);
    o[NTSZ] = 0;
    i = NTSZ - 1;
    do {
        o[i] = table[num % size];
        num /= size;
    } while(num && (--i >= 0));
    if(size == 2)  i = 0;   // all need zeroes
    if(size == 4)  i = NTSZ - (NTSZ / 2);
    if(size == 8)  i = NTSZ - ((NTSZ / 3) + 1);
    if(size == 16) i = NTSZ - (NTSZ / 4);
    if(sign) o[--i] = '-';
    return(o + i);
}


